/*
  Tool            sort_test
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE 4.50 rc 17 upwards, all variants
  Version         1.3.1   5 Jan 2024

  This tool tests the sort macro and TSE's internal and external sort.

  This is done by
  - calling TSE's sort macro explicitly for TSE's internal and external sort,
  - doing an independent check of each sort result,
  - comparing if the internal and external sort produce an identical result.

  Each column sort is also checked for "stability".
    https://en.wikipedia.org/wiki/Sorting_algorithm#Stability

  This "sort_test" tool requires TSE v4.50 rc 17 upwards.
  It works for all 3 TSE variants: Windows GUI and Console, as well as Linux.


  INSTALLATION

  Just compile and execute the macro.

  There is no configuration and there are no parameters.
  It creates its own test data, for some tests randomly.



  HISTORY

  v1       4 Dec 2023
    Initial release, testing with 2 test sets.

  v1.1     7 Dec 2023
    Rewrote the after-sort testing a lot.
      It now is more robust and extensive, and has less duplicate code.
      It now completely leans on TSE's CompareLines() statement, and assumes
      that the next TSE version fixes its non-standard IGNORE_CASE parameter.

  v1.2    13 Dec 2023
    Added more sort tests.
      In Windows the "reverse column sort with some duplicates" curently fails
      because the internal and external sort differ:
      The internal sort is stable and the external sorts are not.
      tsort is not stable when reverse sorting.
      Linux sort is not stable for normal and reverse sorting.
      I am thinking on what to do.

  v1.3    19 Dec 2023
    Added an explicit stability test to column sorts, where a sort is "stable"
    if lines with the same sort-key maintain their old order:
      https://en.wikipedia.org/wiki/Sorting_algorithm#Stability

  v1.3.1   5 Jan 2024
    Updated the documentation.
    Updated the version requirement to needing at least TSE v4.50 rc 17.

*/





/*
  T E C H N I C A L   B A C K G R O U N D   I N F O

  As of TSE 4.50 rc 14 the internal Sort()'s maximum key length is 283
  for both Windows and Linux TSE. Before that it was 287 in Linux TSE.
    https://groups.google.com/g/semware/c/EpdCpV_DNvs/m/px5n1c6IAAAJ
    https://groups.google.com/g/semware/c/EyARyZMRPSQ
*/





// Compatibility check

#ifndef INTERNAL_VERSION
  #define INTERNAL_VERSION 0
#endif

#if INTERNAL_VERSION < 12379

  Version error: You need at least TSE v4.50 rc 17 for this macro.

#endif





// Global constants and semi-constants.
string MACRO_NAME [MAXSTRINGLEN] = ''

#if INTERNAL_VERSION > 12375
  #define IGNORE_CASE _IGNORE_CASE_
#else
  #define IGNORE_CASE 1
#endif


// Global variables.
integer compare_columns_id = 0
integer fingerprint_id     = 0
integer log_id             = 0


// For debugging:
proc log_current_buffer_to_tmp_dir(string identifier)
  integer time = GetTime()
  SaveAs(Format(AddTrailingSlash(GetEnvStr('TMP')),
                time, '_', identifier, '.txt'),
         _DONT_PROMPT_|_OVERWRITE_)
  while time == GetTime()
    Delay(1)
  endwhile
end log_current_buffer_to_tmp_dir


/*
  Based on v0.4 of the fingerprint implementation.
*/
string proc get_buffer_fingerprint(integer buffer_id)
  integer block_value         = 0
  integer byte_count          = 0
  integer byte_gig_count      = 0
  integer byte_position       = 0
  integer byte_value          = 0
  integer crc_value           = 0
  integer fletcher_1st_sum    = 0
  integer fletcher_2nd_sum    = 0
  string  result         [34] = ''

  PushLocation()

  if GotoBufferId(buffer_id)
    BegFile()
    repeat
      byte_value = CurrChar()
      if  byte_value == _AT_EOL_
      and not BinaryMode()
        byte_value = 10
      endif
      if byte_value <> _AT_EOL_
        byte_count    = byte_count + 1
        byte_position = byte_position + 1
        block_value   = block_value * 256 + byte_value
        if byte_count == 1000000000
          byte_count = 0
          byte_gig_count = iif(byte_gig_count == 9, 5, byte_gig_count + 1)
        endif
        if byte_position == 3
          crc_value        = crc_value ^ block_value
          fletcher_1st_sum = (fletcher_1st_sum + block_value     ) mod 16777215
          fletcher_2nd_sum = (fletcher_2nd_sum + fletcher_1st_sum) mod 16777215
          block_value      = 0
          byte_position    = 0
        endif
      endif
    until not NextChar()
    if byte_position <> 0
      crc_value        = crc_value ^ block_value
      fletcher_1st_sum = (fletcher_1st_sum + block_value     ) mod 16777215
      fletcher_2nd_sum = (fletcher_2nd_sum + fletcher_1st_sum) mod 16777215
    endif
    result = Format(fletcher_2nd_sum:8:'0',
                    fletcher_1st_sum:8:'0',
                    crc_value:8:'0',
                    byte_gig_count,
                    byte_count:9:'0')
  endif

  PopLocation()
  return(result)
end get_buffer_fingerprint


integer proc compare_lines(integer line_1,
                           integer line_2,
                           integer sort_flags)
  integer result = FALSE
  integer flag   = _DEFAULT_

  flag   = iif(sort_flags & _IGNORE_CASE_, IGNORE_CASE, _DEFAULT_)
  result = CompareLines(line_1, line_2, flag)
  result = result * iif(sort_flags & _DESCENDING_, -1, 1)

  return(result)
end compare_lines


integer proc compare_columns_7(integer line_1,
                               integer line_2,
                               integer column_from,
                               integer column_to,
                               integer sort_flags,
                               integer buffer_1,
                               integer buffer_2)
  integer old_UnMarkAfterPaste = Set(UnMarkAfterPaste, ON)
  integer result               = 0

  PushLocation()
  PushBlock()

  GotoBufferId(buffer_1)
  MarkColumn(line_1, column_from, line_1, column_to)

  GotoBufferId(compare_columns_id)
  EmptyBuffer()
  CopyBlock()

  GotoBufferId(buffer_2)
  MarkColumn(line_2, column_from, line_2, column_to)

  GotoBufferId(compare_columns_id)
  AddLine()
  BegLine()
  CopyBlock()

  result = compare_lines(1, 2, sort_flags)

  PopLocation()
  PopBlock()
  Set(UnMarkAfterPaste, old_UnMarkAfterPaste)
  return(result)
end compare_columns_7


integer proc compare_columns_5(integer line_1,
                               integer line_2,
                               integer column_from,
                               integer column_to,
                               integer sort_flags)
  integer result = 0

  result = compare_columns_7(line_1,
                             line_2,
                             column_from,
                             column_to,
                             sort_flags,
                             GetBufferId(),
                             GetBufferId())
  return(result)
end compare_columns_5


integer proc test_sorted(integer sort_id,
                         integer line_from,
                         integer column_from,
                         integer line_to,
                         integer column_to,
                         integer sort_type,
                         integer sort_flags)
  integer ok        = TRUE

  if  ok
  and line_from < 1
    Warn('Column line-from is zero or negative.')
    ok = FALSE
  endif

  if ok
    GotoBufferId(sort_id)
    GotoLine(line_from)

    case sort_type
      when _LINE_
        while CurrLine() < line_to
        and   compare_lines(CurrLine()    ,
                           CurrLine() + 1,
                           sort_flags    ) <= 0
          Down()
        endwhile
      when _COLUMN_
        while CurrLine() < line_to
        and   compare_columns_5(CurrLine()    ,
                                CurrLine() + 1,
                                column_from   ,
                                column_to     ,
                                sort_flags    ) <= 0
          Down()
        endwhile
      otherwise
        Warn('ERROR: Invalid sort type.')
        ok = FALSE
    endcase

    if  ok
    and CurrLine() < line_to
      Down()
      MarkLine(CurrLine(), CurrLine())
      UpdateDisplay(_ALL_WINDOWS_REFRESH_)
      Warn('This line is incorrectly sorted!')
      ok = FALSE
    endif
  endif

  return(ok)
end test_sorted


string proc fingerprint_sorted_key_matches(    integer sort_id,
                                               integer stability_line_from,
                                           var integer stability_line_to,
                                               integer line_from,
                                               integer column_from,
                                               integer line_to,
                                               integer column_to,
                                               integer sort_flags)
  string  fingerprint          [34] = ''
  integer old_InsertLineBlocksAbove = Set(InsertLineBlocksAbove, OFF)
  integer old_UnMarkAfterPaste      = Set(UnMarkAfterPaste     , ON)

  if line_from <> line_from endif   // Pacify compiler

  EmptyBuffer(fingerprint_id)

  GotoBufferId(sort_id)
  GotoLine(stability_line_from)

  if stability_line_from <= line_to
    MarkLine(stability_line_from, stability_line_to) // The first two lines.
    GotoBufferId(fingerprint_id)
    CopyBlock()
    EndFile()

    GotoBufferId(sort_id)
    GotoLine(stability_line_to)

    while CurrLine() < line_to
    and   compare_columns_5(CurrLine()    ,
                            CurrLine() + 1,
                            column_from   ,
                            column_to     ,
                            sort_flags    ) == 0
      Down()
      stability_line_to = CurrLine()
      MarkLine(stability_line_to, stability_line_to)
      GotoBufferId(fingerprint_id)
      CopyBlock()
      EndFile()
      GotoBufferId(sort_id)
    endwhile
  endif

  fingerprint = get_buffer_fingerprint(fingerprint_id)

  Set(InsertLineBlocksAbove, old_InsertLineBlocksAbove)
  Set(UnMarkAfterPaste     , old_UnMarkAfterPaste)
  return(fingerprint)
end fingerprint_sorted_key_matches


string proc fingerprint_unsorted_key_matches(integer sort_id,
                                             integer reference_line,
                                             integer unsorted_id,
                                             integer line_from,
                                             integer column_from,
                                             integer line_to,
                                             integer column_to,
                                             integer sort_flags)
  string  fingerprint          [34] = ''
  integer is_small_column           = TRUE
  string  search_key [MAXSTRINGLEN] = ''
  string  search_options        [3] = ''
  integer old_InsertLineBlocksAbove = Set(InsertLineBlocksAbove, OFF)
  integer old_UnMarkAfterPaste      = Set(UnMarkAfterPaste     , ON)

  is_small_column = (column_to + 1 - column_from <= MAXSTRINGLEN)
  search_options  = 'gl' + iif(sort_flags & _IGNORE_CASE_, 'i', '')
  EmptyBuffer(fingerprint_id)
  PushLocation()

  GotoBufferId(sort_id)
  PushLocation()
  GotoLine(reference_line)
  search_key = GetText(column_from, column_to + 1 - column_from)
  PopLocation()

  GotoBufferId(unsorted_id)
  PushLocation()
  PushBlock()
  MarkColumn(line_from, column_from, line_to, column_to)

  while lFind(search_key, search_options)
    if is_small_column
    or compare_columns_7(CurrLine()    ,
                         reference_line,
                         column_from   ,
                         column_to     ,
                         sort_flags    ,
                         GetBufferId() ,
                         sort_id       ) == 0
      PushBlock()
      MarkLine(CurrLine(), CurrLine())
      GotoBufferId(fingerprint_id)
      CopyBlock()
      EndFile()
      GotoBufferId(unsorted_id)
      PopBlock()
    endif

    EndLine()
    search_options[1] = '+'
  endwhile

  fingerprint = get_buffer_fingerprint(fingerprint_id)

  PopBlock()
  PopLocation()
  PopLocation()
  Set(InsertLineBlocksAbove, old_InsertLineBlocksAbove)
  Set(UnMarkAfterPaste     , old_UnMarkAfterPaste)
  return(fingerprint)
end fingerprint_unsorted_key_matches


string proc test_stability(integer sort_id,
                           integer unsorted_id,
                           integer line_from,
                           integer column_from,
                           integer line_to,
                           integer column_to,
                           integer sort_flags)
  string  fingerprint_sorted   [34] = ''
  integer stability_line_from       = 0
  integer stability_line_to         = 0
  string  fingerprint_unsorted [34] = ''
  string  unstable_lines       [25] = ''

  if line_from < line_to
    GotoBufferId(sort_id)
    GotoLine(line_from)

    while unstable_lines == ''
    and   CurrLine() < line_to
    and   Down()
      if compare_columns_5(CurrLine() - 1,
                           CurrLine(),
                           column_from,
                           column_to,
                           sort_flags) == 0  // The lines' sort keys are equal.
        stability_line_from  = CurrLine() - 1
        stability_line_to    = CurrLine() // The proc updates it if more lines.
        fingerprint_sorted   = fingerprint_sorted_key_matches  (sort_id,
                                                                stability_line_from,
                                                                stability_line_to,
                                                                line_from,
                                                                column_from,
                                                                line_to,
                                                                column_to,
                                                                sort_flags)
        fingerprint_unsorted = fingerprint_unsorted_key_matches(sort_id,
                                                                stability_line_from,
                                                                unsorted_id,
                                                                line_from,
                                                                column_from,
                                                                line_to,
                                                                column_to,
                                                                sort_flags)
        if fingerprint_sorted <> fingerprint_unsorted
          unstable_lines = Format(stability_line_from; '-'; stability_line_to)
        endif
      endif
    endwhile
  endif

  return(unstable_lines)
end test_stability


integer proc display_test_result(integer old_MsgLevel,
                                 string  test_title,
                                 integer internal_sort_ok,
                                 integer external_sort_ok,
                                 integer unsorted_id,
                                 integer internal_sort_id,
                                 integer external_sort_id,
                                 integer line_from,
                                 integer column_from,
                                 integer line_to,
                                 integer column_to,
                                 integer sort_type,
                                 integer sort_flags)
  string  external_sort_fingerprint [34] = ''
  string  internal_sort_fingerprint [34] = ''
  integer ok                             = TRUE
  string  unstable_lines            [25] = ''

  Set(MsgLevel, old_MsgLevel)

  if     not internal_sort_ok
    AddLine(Format('Test "', test_title,'" failed: Internal sort failed.'),
            log_id)
    ok = FALSE
  elseif not external_sort_ok
    AddLine(Format('Test "', test_title,'" failed: External sort failed.'),
            log_id)
    ok = FALSE
  endif

  if ok
    if test_sorted(internal_sort_id,
                   line_from,
                   column_from,
                   line_to,
                   column_to,
                   sort_type,
                   sort_flags)
      if sort_type & _COLUMN_
        unstable_lines = test_stability(internal_sort_id,
                                        unsorted_id,
                                        line_from,
                                        column_from,
                                        line_to,
                                        column_to,
                                        sort_flags)
        if unstable_lines == ''
          AddLine(Format('Test "', test_title,
                         '" note: Internal buffer sort was stable.'),
                  log_id)
        else
          AddLine(Format('Test "', test_title,
                         '" note: Internal buffer sort not stable for sorted lines';
                         unstable_lines),
                  log_id)
          ok = FALSE  // xxxx
        endif
      endif
    else
      AddLine(Format('Test "', test_title,
                     '" failed: Internal buffer not sorted correctly.'),
              log_id)
      ok = FALSE
    endif

    if test_sorted(external_sort_id,
                   line_from,
                   column_from,
                   line_to,
                   column_to,
                   sort_type,
                   sort_flags)
      if sort_type & _COLUMN_
        unstable_lines = test_stability(external_sort_id,
                                        unsorted_id,
                                        line_from,
                                        column_from,
                                        line_to,
                                        column_to,
                                        sort_flags)
        if unstable_lines == ''
          AddLine(Format('Test "', test_title,
                         '" note: External buffer sort was stable.'),
                  log_id)
        else
          AddLine(Format('Test "', test_title,
                         '" note: External buffer sort not stable for sorted lines';
                         unstable_lines),
                  log_id)
        endif
      endif
    else
      AddLine(Format('Test "', test_title,
                     '" failed: External buffer not sorted correctly.'),
              log_id)
      ok = FALSE
    endif
  endif

  if ok
    internal_sort_fingerprint = get_buffer_fingerprint(internal_sort_id)
    external_sort_fingerprint = get_buffer_fingerprint(external_sort_id)

    if internal_sort_fingerprint <> external_sort_fingerprint
      if sort_type & _COLUMN_
        AddLine(Format('Test "', test_title,
                       '" warning: Internal and external sort differ.'),
                log_id)
      else
        AddLine(Format('Test "', test_title,
                       '" error: Internal and external sort differ.'),
                log_id)
        ok = FALSE
      endif
    endif
  endif

  if ok
    AddLine(Format('Test "', test_title,'" OK.'), log_id)
  endif

  return(ok)
end display_test_result


integer proc display_test_start(string test_title)
  integer old_MsgLevel = 0

  Message('Testing "', test_title, '" ... ')

  old_MsgLevel = Set(MsgLevel, _NONE_)
  return(old_MsgLevel)
end display_test_start


integer proc exec_macro(string cmd)
  integer ok           = FALSE
  integer old_MsgLevel = Query(MsgLevel)
  #ifdef LINUX
    #if INTERNAL_VERSION < 12350  // < TSE v4.46
      // In TSE v4.46 downwards the Linux ExecMacro() command erroneously
      // returns TRUE if the called macro's .mac does not exist.
      // Here we fix that, faithfully reproducing the warning.
      if search_path(GetToken(cmd, ' ', 1) + '.mac', '', 'mac') == ''
        if Query(MsgLevel) <> _NONE_
          Warn('File not found:'; GetToken(cmd, ' ', 1) + '.mac')
        endif
        return(FALSE)
      endif
    #endif
  #endif
  Set(MsgLevel, _NONE_)
  ok = ExecMacro(cmd)
  Set(MsgLevel, old_MsgLevel)
  if not ok
    Warn(MACRO_NAME; 'error Macro Executing "', cmd, '".')
  endif
  return(ok)
end exec_macro


proc copy_buffer(integer from_id, integer to_id)
  GotoBufferId(from_id)
  MarkAll()
  GotoBufferId(to_id)
  SetUndoOff()
  EmptyBuffer()
  CopyBlock()
end copy_buffer


proc make_test_set_1(integer buffer_id)
  string  char [1] = ''
  integer col      = 0
  integer line     = 0

  GotoBufferId(buffer_id)
  SetUndoOff()
  EmptyBuffer()

  for line = 1 to 10000
    AddLine()
    EndLine()
    char = 'a'
    for col = 1 to 283 + 1 - (line mod 283 + 1)
      InsertText(char)
      if char == 'z'
        char = 'a'
      else
        char = Chr(Asc(char) + 1)
      endif
    endfor
  endfor
end make_test_set_1


proc make_test_set_2(integer buffer_id)
  integer line   = 0
  integer column = 0

  GotoBufferId(buffer_id)
  SetUndoOff()
  EmptyBuffer()

  for line = 1 to 10000
    AddLine()
    BegLine()
    for column = 1 to 283
      InsertText(Chr(Random(65, 90)))
    endfor
  endfor
end make_test_set_2


proc make_test_set_3(integer buffer_id)
  integer line   = 0
  integer column = 0

  GotoBufferId(buffer_id)
  SetUndoOff()
  EmptyBuffer()

  for line = 1 to 10000
    AddLine()
    BegLine()
    for column = 1 to 20
      InsertText(Chr(Random(65, 90)))
    endfor
    if  Random(1, 10) == 1
    and CurrLine() > 1
       MarkColumn(CurrLine() - 1, 21, CurrLine() - 1, 60)
       CopyBlock()
       EndLine()
    else
      for column = 21 to 60
        InsertText(Chr(Random(65, 90)))
      endfor
    endif
    for column = 61 to 300
      InsertText(Chr(Random(65, 90)))
    endfor
  endfor
end make_test_set_3


proc WhenPurged()
end WhenPurged


proc WhenLoaded()
  MACRO_NAME = SplitPath(CurrMacroFilename(), _NAME_)
end WhenLoaded


proc Main()
  integer external_sort_id          = 0
  integer external_sort_ok          = TRUE
  integer internal_sort_id          = 0
  integer internal_sort_ok          = TRUE
  integer ok                        = TRUE
  integer old_MsgLevel              = 0
  string  run_id               [10] = ''
  string  test_title [MAXSTRINGLEN] = ''
  integer unsorted_id               = 0

  // SetRandomSeed(123456789)  // Same random test-data every run.

  PushLocation()
  log_id = NewFile()
  run_id = StrReplace('^[~0-9]#{[0-9]#}.#$', CurrFilename(), '\1', 'x')
  ChangeCurrFilename(MACRO_NAME + ':' + run_id + ':Log',
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
  unsorted_id        = NewFile()
  ChangeCurrFilename(MACRO_NAME + ':' + run_id + ':Unsorted',
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
  internal_sort_id   = NewFile()
  ChangeCurrFilename(MACRO_NAME + ':' + run_id + ':InternallySorted',
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
  external_sort_id   = NewFile()
  ChangeCurrFilename(MACRO_NAME + ':' + run_id + ':ExternallySorted',
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
  compare_columns_id = CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + ':' + run_id + ':Compare_Columns',
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
  fingerprint_id     = CreateTempBuffer()
  ChangeCurrFilename(MACRO_NAME + ':' + run_id + ':Fingerprint',
                     _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)


  if ok
    test_title   = 'Sort Maximally Reversely Sorted Buffer'
    old_MsgLevel = display_test_start(test_title)

    make_test_set_1(unsorted_id)
    log_current_buffer_to_tmp_dir('SortBefore')

    copy_buffer(unsorted_id, internal_sort_id)
    copy_buffer(unsorted_id, external_sort_id)

    GotoBufferId(internal_sort_id)
    log_current_buffer_to_tmp_dir('InternalSortBefore')
    MarkAll()
    internal_sort_ok = exec_macro('sort --internal')
    log_current_buffer_to_tmp_dir('InternalSortAfter')

    GotoBufferId(external_sort_id)
    log_current_buffer_to_tmp_dir('ExternalSortBefore')
    MarkAll()
    external_sort_ok = exec_macro('sort --external')
    log_current_buffer_to_tmp_dir('ExternalSortAfter')

    ok = display_test_result(old_MsgLevel, test_title,
                            internal_sort_ok, external_sort_ok,
                            unsorted_id, internal_sort_id, external_sort_id,
                            // line from, column from, line to, column to
                            1, _NONE_, NumLines(), _NONE_,
                            // sort type, sort flags
                            _LINE_, _DEFAULT_)
  endif

  if ok
    test_title   = 'Reverse Sorting a Sorted Buffer'
    old_MsgLevel = display_test_start(test_title)

    GotoBufferId(internal_sort_id)
    log_current_buffer_to_tmp_dir('ReverseInternalSortBefore')
    MarkAll()
    internal_sort_ok = exec_macro('sort --internal -d')
    log_current_buffer_to_tmp_dir('ReverseInternalSortAfter')

    GotoBufferId(external_sort_id)
    log_current_buffer_to_tmp_dir('ReverseExternalSortBefore')
    MarkAll()
    external_sort_ok = exec_macro('sort --external -d')
    log_current_buffer_to_tmp_dir('ReverseExternalSortAfter')

    ok = display_test_result(old_MsgLevel, test_title,
                            internal_sort_ok, external_sort_ok,
                            unsorted_id, internal_sort_id, external_sort_id,
                            // line from, column from, line to, column to
                            1, _NONE_, NumLines(), _NONE_,
                            // sort type, sort flags
                            _LINE_, _DESCENDING_)
  endif


  if ok
    test_title   = 'Sort Completely Random Buffer'
    old_MsgLevel = display_test_start(test_title)

    make_test_set_2(unsorted_id)
    log_current_buffer_to_tmp_dir('RandomBufferBefore')
    copy_buffer(unsorted_id, internal_sort_id)
    copy_buffer(unsorted_id, external_sort_id)

    GotoBufferId(internal_sort_id)
    log_current_buffer_to_tmp_dir('RandomInternalSortBefore')
    MarkAll()
    internal_sort_ok = exec_macro('sort --internal')
    log_current_buffer_to_tmp_dir('RandomInternalSortAfter')

    GotoBufferId(external_sort_id)
    log_current_buffer_to_tmp_dir('RandomExternalSortBefore')
    MarkAll()
    external_sort_ok = exec_macro('sort --external')
    log_current_buffer_to_tmp_dir('RandomExternalSortAfter')

    ok = display_test_result(old_MsgLevel, test_title,
                            internal_sort_ok, external_sort_ok,
                            unsorted_id, internal_sort_id, external_sort_id,
                            // line from, column from, line to, column to
                            1, _NONE_, NumLines(), _NONE_,
                            // sort type, sort flags
                            _LINE_, _DEFAULT_)
  endif


  if ok
    test_title   = 'Reverse Sort Completely Random Buffer'
    old_MsgLevel = display_test_start(test_title)

    make_test_set_2(unsorted_id)
    copy_buffer(unsorted_id, internal_sort_id)
    copy_buffer(unsorted_id, external_sort_id)

    GotoBufferId(internal_sort_id)
    MarkAll()
    internal_sort_ok = exec_macro('sort --internal -d')

    GotoBufferId(external_sort_id)
    MarkAll()
    external_sort_ok = exec_macro('sort --external -d')

    ok = display_test_result(old_MsgLevel, test_title,
                            internal_sort_ok, external_sort_ok,
                            unsorted_id, internal_sort_id, external_sort_id,
                            // linemc from, column from, line to, column to
                            1, _NONE_, NumLines(), _NONE_,
                            // sort type, sort flags
                            _LINE_, _DESCENDING_)
  endif


  if ok
    test_title   = 'Sort Random Columns with Some Duplicates'
    old_MsgLevel = display_test_start(test_title)

    make_test_set_3(unsorted_id)
    copy_buffer(unsorted_id, internal_sort_id)
    copy_buffer(unsorted_id, external_sort_id)

    GotoBufferId(internal_sort_id)
    MarkColumn(20, 21, 10000 - 20, 60)
    internal_sort_ok = exec_macro('sort --internal')

    GotoBufferId(external_sort_id)
    MarkColumn(20, 21, 10000 - 20, 60)
    external_sort_ok = exec_macro('sort --external')

    ok = display_test_result(old_MsgLevel, test_title,
                            internal_sort_ok, external_sort_ok,
                            unsorted_id, internal_sort_id, external_sort_id,
                            // line from, column from, line to, column to
                            20, 21, 10000 - 20, 60,
                            // sort type, sort flags
                            _COLUMN_, _DEFAULT_)
  endif


  if ok
    test_title   = 'Reverse Sort Random Columns with Some Duplicates'
    old_MsgLevel = display_test_start(test_title)

    make_test_set_3(unsorted_id)
    copy_buffer(unsorted_id, internal_sort_id)
    copy_buffer(unsorted_id, external_sort_id)

    GotoBufferId(internal_sort_id)
    MarkColumn(20, 21, 10000 - 20, 60)
    internal_sort_ok = exec_macro('sort --internal -d')

    GotoBufferId(external_sort_id)
    MarkColumn(20, 21, 10000 - 20, 60)
    external_sort_ok = exec_macro('sort --external -d')

    ok = display_test_result(old_MsgLevel, test_title,
                            internal_sort_ok, external_sort_ok,
                            unsorted_id, internal_sort_id, external_sort_id,
                            // line from, column from, line to, column to
                            20, 21, 10000 - 20, 60,
                            // sort type, sort flags
                            _COLUMN_, _DESCENDING_)
  endif


  if ok
    PopLocation()
    GotoBufferId(log_id)
    AbandonFile(unsorted_id)
    AbandonFile(internal_sort_id)
    AbandonFile(external_sort_id)
    AbandonFile(compare_columns_id)
    AbandonFile(fingerprint_id)
  else
    KillLocation()
    GotoBufferId(log_id)
    if Query(Beep)
      Alarm()
    endif
  endif

  PurgeMacro(CurrMacroFilename())
end Main

